/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


package de.tudarmstadt.maki.simonstrator.api.component.sis.prediction;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import de.tudarmstadt.maki.simonstrator.api.Event;
import de.tudarmstadt.maki.simonstrator.api.EventHandler;
import de.tudarmstadt.maki.simonstrator.api.Host;
import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.common.graph.INodeID;
import de.tudarmstadt.maki.simonstrator.api.component.ComponentNotAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSComponent;
import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSDataCallback;
import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSInfoProperties;
import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSInformationProvider.SiSProviderHandle;
import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSRequest;
import de.tudarmstadt.maki.simonstrator.api.component.sis.exception.InformationNotAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.sis.type.PredictionContainer;
import de.tudarmstadt.maki.simonstrator.api.component.sis.type.PredictionSiSType;
import de.tudarmstadt.maki.simonstrator.api.component.sis.type.SiSType;

/**
 * @author Tobias Meuser (tobias.meuser@kom.tu-darmstadt.de)
 * @version 1.0 at 03.01.2018
 *
 */
public class DefaultSiSPredictionManager implements SiSPredictionManager, EventHandler {

	private Map<SiSType<? extends Object>, Prediction<? extends Object>> availablePredictions = new HashMap<>();

	private Map<SiSType<?extends Object>,
			Map<INodeID, SiSValueContainer<? extends Object>>> availableData = new HashMap<>();


	private Host host;

	private SiSComponent _siSComponent;

	public DefaultSiSPredictionManager(Host host, List<Prediction<? extends Object>> pPredictions) {
		this.host = host;
		
		for (Prediction<? extends Object> prediction : pPredictions) {
			availablePredictions.put(prediction.getBasicSiSType(), prediction);
		}
	}

	@Override
	public void initialize() {
		try {
			_siSComponent = getHost().getComponent(SiSComponent.class);

			for (Entry<SiSType<? extends Object>, Prediction<? extends Object>> entry : availablePredictions
					.entrySet()) {
				PredictionSiSType<? extends Object> predictionSysType = entry.getKey().getPredictionSysType();
				if (predictionSysType != null) {
					_siSComponent.provide().nodeState(predictionSysType,
							new PredictionSiSDataCallback<>(predictionSysType));
				}

				availableData.put(entry.getKey(), new HashMap<>());
			}

			Event.scheduleWithDelay(Time.SECOND, this, null, 0);
		} catch (ComponentNotAvailableException e) {
		}
	}

	@Override
	public void shutdown() {

	}

	@Override
	public Host getHost() {
		return host;
	}

	@Override
	public <T> PredictionContainer<T> predict(INodeID pNodeID, PredictionSiSType<T> pType, T pValue, long pTime)
			throws InformationNotAvailableException {
		if (availablePredictions.containsKey(pType.getOriginalSiSType())) {
			Prediction prediction = availablePredictions.get(pType.getOriginalSiSType());
			return prediction.predict(pNodeID, pValue, pTime);
		}
		throw new InformationNotAvailableException();
	}

	@Override
	public <T> PredictionContainer<T> predict(INodeID pNodeID, PredictionSiSType<T> pType, T pValue, long pTime,
			long pPredictionTimestamp) throws InformationNotAvailableException {
		if (availablePredictions.containsKey(pType.getOriginalSiSType())) {
			Prediction prediction = availablePredictions.get(pType.getOriginalSiSType());
			return prediction.predict(pNodeID, pValue, pTime, pPredictionTimestamp);
		}
		throw new InformationNotAvailableException();
	}

	@SuppressWarnings("rawtypes")
	private class PredictionSiSDataCallback<T extends Object> implements SiSDataCallback<PredictionContainer> {

		private PredictionSiSType<T> _sisType;

		public PredictionSiSDataCallback(PredictionSiSType<T> sisType) {
			_sisType = sisType;
		}

		@SuppressWarnings("unchecked")
		@Override
		public PredictionContainer<T> getValue(INodeID pNodeID, SiSProviderHandle pProviderHandle)
				throws InformationNotAvailableException {

			SiSValueContainer<? extends Object> siSValueContainer = availableData.get(_sisType.getOriginalSiSType())
					.get(pNodeID);

			if (siSValueContainer != null) {
				T observation = (T) siSValueContainer.getValue();
				long timestamp = siSValueContainer.getTimestamp();

				return predict(pNodeID, _sisType, observation, timestamp);
			} else {
				throw new InformationNotAvailableException();
			}
		}

		@Override
		public Set<INodeID> getObservedNodes() {
			return _siSComponent.get().getObservedNodes(_sisType.getOriginalSiSType(), SiSRequest.NONE);
		}

		@Override
		public SiSInfoProperties getInfoProperties() {
			return new SiSInfoProperties();
		}
		
	}

	@Override
	public void updateValue(SiSType<Object> type, INodeID id, Object value, long timestamp) {
		if (availableData.containsKey(type)) {
			availableData.get(type).put(id, new SiSValueContainer<Object>(value, timestamp));
		}
	}

	@Override
	public void eventOccurred(Object pContent, int pType) {
		for (SiSType<? extends Object> type : availableData.keySet()) {
			Map<INodeID, SiSValueContainer<? extends Object>> nodeValues = availableData.get(type);
			Set<INodeID> observedNodes = _siSComponent.get().getObservedNodes(type, SiSRequest.NONE);

			for (INodeID nodeID : observedNodes) {
				try {
					Object measurement = _siSComponent.get().localObservationOf(nodeID, type, SiSRequest.NONE);
					if (nodeValues.containsKey(nodeID)) {
						if (!nodeValues.get(nodeID).getValue().equals(measurement)) {
							nodeValues.put(nodeID, new SiSValueContainer<Object>(measurement, Time.getCurrentTime()));
						}
					} else {
						nodeValues.put(nodeID, new SiSValueContainer<Object>(measurement, Time.getCurrentTime()));
					}
				} catch (InformationNotAvailableException e) {
				}
			}
		}

		Event.scheduleWithDelay(Time.SECOND, this, null, 0);
	}
	
	private class SiSValueContainer<T extends Object> {
		private T value;
		private long timestamp;

		public SiSValueContainer(T pValue, long pTimestamp) {
			value = pValue;
			timestamp = pTimestamp;
		}

		public T getValue() {
			return value;
		}

		public long getTimestamp() {
			return timestamp;
		}

		/**
		 * @param pTimestamp
		 *            the timestamp to set
		 */
		public void setTimestamp(long pTimestamp) {
			timestamp = pTimestamp;
		}
	}
}
